/*
 * Copyright (c) 2012, 2015, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.oracle.truffle.r.nodes.query;

import qir.ast.QIRNode;

import static com.oracle.truffle.r.runtime.RVisibility.OFF;
import static com.oracle.truffle.r.runtime.builtins.RBehavior.COMPLEX;
import static com.oracle.truffle.r.runtime.builtins.RBuiltinKind.PRIMITIVE;

import com.oracle.truffle.api.CompilerDirectives.TruffleBoundary;
import com.oracle.truffle.api.dsl.*;
import com.oracle.truffle.r.nodes.builtin.RBuiltinNode;
import com.oracle.truffle.r.nodes.qirinterface.QIRInterface;
import com.oracle.truffle.r.runtime.builtins.RBuiltin;
import com.oracle.truffle.r.runtime.context.RContext;
import com.oracle.truffle.r.runtime.data.RDataFactory;
import com.oracle.truffle.r.runtime.data.RList;
import com.oracle.truffle.r.runtime.env.REnvironment;
import com.oracle.truffle.r.runtime.env.REnvironment.PutException;

/**
 * This builtin takes an object returned by a query and returns a tuple from the result. This tuple
 * is an object with a "get" method to fetch a column given the name.
 */
@RBuiltin(name = "query.force", visibility = OFF, kind = PRIMITIVE, parameterNames = {"query"}, behavior = COMPLEX)
public abstract class RNextResQueryBuiltin extends RBuiltinNode {
    @TruffleBoundary
    @Specialization
    public static final RList force(final REnvironment rQuery) {
        // Every query is identified by an integer included in the object returned by the query.
        final int queryId = (int) rQuery.get("queryId");
        if (queryId == RContext.INVALID_QUERY_ID)
            throw new RuntimeException("Not a query object.");

        // TODO: Handle results given in streaming.
        QIRNode query = RContext.queries.get(queryId);
        if (query == null)
            throw new RuntimeException("Unrecognized or empty query: " + queryId);

        RList queryResults = RContext.results.get(queryId);
        if (queryResults != null) // The query has been run before.
            return queryResults;
        query = QIRInterface.normalize(query, RContext.envs.get(queryId));
        RContext.queries.set(queryId, query);

        queryResults = RDataFactory.createList(QIRInterface.run(query).stream().map(h -> {
            final REnvironment env = RDataFactory.createNewEnv(null);
            h.entrySet().stream().forEach(entry -> {
                try {
                    env.put(entry.getKey(), entry.getValue());
                } catch (PutException e) {
                    e.printStackTrace();
                }
            });
            return env;
        }).toArray());
        RContext.results.set(queryId, queryResults);

        return queryResults;
    }
}