/*
 * Copyright (c) 2015, 2015, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.oracle.truffle.r.nodes.qirinterface;

import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import com.oracle.truffle.api.frame.Frame;
import com.oracle.truffle.api.frame.FrameSlot;
import com.oracle.truffle.api.source.Source;
import com.oracle.truffle.api.source.SourceSection;
import com.oracle.truffle.r.nodes.function.FunctionDefinitionNode;
import com.oracle.truffle.r.nodes.function.FunctionExpressionNode;
import com.oracle.truffle.r.runtime.RRuntime;
import com.oracle.truffle.r.runtime.context.RContext;
import com.oracle.truffle.r.runtime.data.RFunction;
import com.oracle.truffle.r.runtime.env.REnvironment;

import qir.ast.*;
import qir.ast.data.*;
import qir.ast.value.*;
import qir.driver.sql.OracleDriver;
import qir.driver.sql.PostgreSQLDriver;
import qir.driver.ConnectionData;
import qir.driver.DBDriver;
import qir.driver.QIRDriver;
import qir.driver.hbase.HBaseDriver;

public final class QIRInterface {
    private static final Map<String, DBDriver<?, ?>> driverCache = new HashMap<>();

    /**
     * Runs a query in the database pointed by the given driver and returns the results.
     *
     * @param query The query to be executed in the targeted database
     * @return The results of the query
     */
    public static final ArrayList<Map<String, Object>> run(final QIRNode query) {
        QIRNode qirRes = QIRDriver.run(query);
        if (qirRes == null)
            throw new RuntimeException("Could not run query: " + query);
        final ArrayList<Map<String, Object>> result = new ArrayList<>();
        while (qirRes != QIRLnil.getInstance()) {
            final HashMap<String, Object> tuple = new HashMap<>();
            QIRNode qirTuple = ((QIRLcons) qirRes).value;
            while (qirTuple != QIRTnil.getInstance()) {
                tuple.put(((QIRTcons) qirTuple).id, QIRToRType(((QIRTcons) qirTuple).value));
                qirTuple = ((QIRTcons) qirTuple).tail;
            }
            result.add(tuple);
            qirRes = ((QIRLcons) qirRes).tail;
        }
        return result;
    }

    /**
     * Resolves the free variables of a query.
     *
     * @param arg The query to be executed in the targeted database
     * @param frame The environment of execution
     * @return The closed query
     */
    public static final QIRNode normalize(final QIRNode arg, final Frame frame) {
        QIRNode query = arg;
        final SourceSection dummy = Source.newBuilder("").name("QIR interface").mimeType(RRuntime.R_APP_MIME).build().createUnavailableSection();

        for (Map<String, QIRVariable> fvs = QIRDriver.getFreeVars(query); !fvs.isEmpty(); fvs = QIRDriver.getFreeVars(query))
            for (final QIRVariable fv : fvs.values()) {
                final String varName = fv.getId();
                final FrameSlot varSlot = frame.getFrameDescriptor().findFrameSlot(varName);
                query = new QIRApply(dummy, new QIRLambda(dummy, null, new QIRVariable(dummy, varName, varSlot), query),
                                RToQIRType(fv.sourceSection, varSlot != null ? frame.getValue(varSlot) : RContext.lookupBuiltin(varName)));
            }
        return query;
    }

    /**
     * Returns a {@link DBDriver} of a database using a configuration file. This function uses a
     * cache for the drivers.
     *
     * @param dbName The name of the targeted database
     * @param fileName The name of the configuration file
     * @return A {@link DBDriver} to the given database with the given configuration
     */
    public static final DBDriver<?, ?> createDriver(final String dbName, final String fileName) throws UnsupportedOperationException {
        final String cacheKey = dbName + '@' + fileName;
        DBDriver<?, ?> result = driverCache.get(cacheKey);

        if (result != null)
            return result;
        switch (dbName) {
            case "Oracle":
                result = new OracleDriver(createConnectionData("ora", fileName));
                break;
            case "PostgreSQL":
                result = new PostgreSQLDriver(createConnectionData("pgsql", fileName));
                break;
            case "HBase":
                result = new HBaseDriver(fileName);
                break;
            default:
                throw new RuntimeException("Unknown database: " + dbName);
        }
        driverCache.put(cacheKey, result);
        return result;
    }

    /**
     * This function closes the connection of drivers created with the function
     * {@link #createDriver(String, String)}.
     */
    public static final void closeDrivers() {
        driverCache.values().stream().forEach(driver -> driver.closeConnection());
    }

    /**
     * Creates a {@link ConnectionData} object based on the given configuration file.
     *
     * @param prefix A prefix for properties in the configuration file
     * @param fileName The name of the configuration file
     * @return A {@link ConnectionData} created from the given configuration file
     */
    private static final ConnectionData createConnectionData(final String prefix, final String fileName) {
        try {
            final Reader dbConfigReader = new FileReader(fileName);
            final Properties p = new Properties();
            p.load(dbConfigReader);
            final String sessionId = p.getProperty(prefix + ".sid");
            final String serverName = p.getProperty(prefix + ".host");
            final int port = Integer.parseInt(p.getProperty(prefix + ".port"));
            final String userName = p.getProperty(prefix + ".user");
            final String passwd = p.getProperty(prefix + ".passwd");
            return new ConnectionData(sessionId, serverName, port, userName, passwd);
        } catch (IOException e) {
            throw new RuntimeException("Invalid configuration file for connection to a database: " + fileName);
        }
    }

    /**
     * Translates a QIR expression into a R statement.
     *
     * @param value The QIR expression to translate
     * @return The translation of the QIR expression in R
     * @throws UnsupportedOperationException If the type of the value is not supported.
     */
    @SuppressWarnings("unchecked")
    private static final <T> T QIRToRType(final QIRNode value) throws UnsupportedOperationException {
        if (value instanceof QIRBaseValue<?>)
            return ((QIRBaseValue<T>) value).value;
        throw new RuntimeException("Unsupported value: " + value);
    }

    /**
     * Translates a R statement into a QIR expression.
     *
     * @param src The {@link SourceSection} of the given value
     * @param value The R statement to translate
     * @return The translation of the R statement in QIR
     */
    static final <T> QIRNode RToQIRType(final SourceSection src, final T value) {
        if (value instanceof BigInteger)
            return new QIRBigNumber(src, (BigInteger) value);
        if (value instanceof Long)
            return new QIRNumber(src, (Long) value);
        if (value instanceof Double)
            return new QIRDouble(src, (Double) value);
        if (value instanceof Boolean)
            return new QIRBoolean(src, (Boolean) value);
        if (value instanceof String)
            return new QIRString(src, (String) value);
        if (value instanceof FunctionExpressionNode) {
            final FunctionExpressionNode fun = (FunctionExpressionNode) value;
            if (fun.getCallTarget() == null)
                return new QIRBuiltin(src, fun.getSyntaxDebugName());
            final String funName = fun.getSyntaxDebugName();
            try {
                QIRNode res = ((FunctionDefinitionNode) fun.getCallTarget().getRootNode()).getBody().accept(QIRTranslateVisitor.instance);
                final String[] args = ((FunctionDefinitionNode) fun.getCallTarget().getRootNode()).getSignature().getNames();
                for (int i = 0; i < args.length; i++)
                    res = new QIRLambda(src, funName, new QIRVariable(null, args[i], null), res);
                return res;
            } catch (UnsupportedOperationException e) {
                final SourceSection funSrc = fun.getCallTarget().getRootNode().getSourceSection();
                final String program = funSrc.getCode() + "\n" + funName;
                // TODO: Handle function dependencies
                return new QIRTruffleNode(funSrc, program);
            }
        }
        if (value instanceof REnvironment) {
            REnvironment env = (REnvironment) value;
            final Object queryId = env.get("queryId");
            if (queryId != null) { // The object is a query
                final int qid = ((Long) queryId).intValue();
                return normalize(RContext.queries.get(qid), RContext.envs.get(qid));
            }

            final String tableName = (String) env.get("tableName");
            final String schemaName = (String) env.get("schemaName");
            final String dbName = (String) env.get("dbName");
            final String configFile = (String) env.get("configFile");
            if (tableName != null && dbName != null && configFile != null) // The object is a table
                return new QIRTable<>(src, new QIRString(null, tableName), new QIRString(null, schemaName), createDriver(dbName, configFile));
            // else, the object is considered a tuple
            QIRTuple tuple = QIRTnil.getInstance();
            for (final Object x : env.getFrame().getFrameDescriptor().getIdentifiers())
                tuple = new QIRTcons(src, (String) x, RToQIRType(src, env.get((String) x)), tuple);
            return tuple;
        }
        if (value instanceof RFunction) {
            final RFunction fun = (RFunction) value;
            if (fun.isBuiltin())
                switch (fun.getName()) {
                    case "new.env":
                        return new QIRLambda(src, "new.env", new QIRVariable(null, "_", null), QIRTnil.getInstance());
                    case "return":
                    case "(": {
                        final QIRVariable x = new QIRVariable(null, "x", null);
                        return new QIRLambda(src, "identity", x, x);
                    }
                    case "c": {
                        // TODO: This works only for lists with one element
                        final QIRVariable x = new QIRVariable(null, "x", null);
                        return new QIRLambda(src, "lcons", x, new QIRLcons(null, x, QIRLnil.getInstance()));
                    }
                    case "sum":
                        return new QIRBuiltin(src, "sum");
                    default:
                        throw new RuntimeException("Unsupported value: " + value + " : " + value.getClass());
                }
            try {
                final FunctionDefinitionNode rootNode = (FunctionDefinitionNode) fun.getRootNode();
                QIRNode res = rootNode.getBody().accept(QIRTranslateVisitor.instance);
                final String[] args = Arrays.asList(rootNode.argSourceSections).stream().map(arg -> arg.getCode()).toArray(size -> new String[size]);
                final String funName = fun.getName();
                if (args.length == 0)
                    return new QIRLambda(src, funName, null, res);
                for (int i = 0; i < args.length; i++)
                    res = new QIRLambda(src, funName, new QIRVariable(null, args[i], null), res);
                return res;
            } catch (UnsupportedOperationException e) {
                final SourceSection funSrc = fun.getRootNode().getSourceSection();
                // TODO: Handle dependencies
                return new QIRTruffleNode(funSrc, funSrc.getCode());
            }
        }
        throw new RuntimeException("Unsupported value: " + value);
    }
}